cimport aho_corasick

cdef aho_corasick.ac_error_code decref_result_object(void *item, void* data):
    aho_corasick.Py_DecRef(<object>item)
    return aho_corasick.AC_SUCCESS

cdef aho_corasick.ac_error_code append_result(void* data,
                                              aho_corasick.ac_result* result):
    result_list = <list>data
    result_tuple = ((result.start, result.end), <object>result.object)
    result_list.append(result_tuple)
    return aho_corasick.AC_SUCCESS

cdef text_as_bytes(text):
    if type(text) is unicode:
        return text.encode('utf8')
    else:
        return text

cdef class Index:
    cdef aho_corasick.ac_index _index
    def __cinit__(self):
        self._index = aho_corasick.ac_index_new()
        if self._index is NULL:
            raise MemoryError()

    def __dealloc__(self):
        aho_corasick.ac_index_free(self._index, decref_result_object)


    def enter(self, keyword, obj=None):
        if obj is None:
            obj = keyword

        keyword = text_as_bytes(keyword)

        if aho_corasick.ac_index_fixed(self._index):
            raise TypeError("Can't call enter after fix")

        if aho_corasick.ac_index_enter(self._index,
                                       keyword, len(keyword),
                                       <void*>obj) != aho_corasick.AC_SUCCESS:
            raise MemoryError()
        aho_corasick.Py_IncRef(obj)

    def fix(self):
        if aho_corasick.ac_index_fixed(self._index):
            raise TypeError("Can't call fix repeatedly")

        if aho_corasick.ac_index_fix(self._index) != aho_corasick.AC_SUCCESS:
            raise MemoryError()

    def query(self, phrase):
        if not aho_corasick.ac_index_fixed(self._index):
            raise TypeError("Can't call query before fix")

        phrase = text_as_bytes(phrase)

        result_list = []

        status = aho_corasick.ac_index_query_cb(self._index,
                                          phrase,
                                          len(phrase),
                                          append_result,
                                          <void*>result_list)
        if status != aho_corasick.AC_SUCCESS:
            raise MemoryError()

        return result_list



#!/usr/bin/env python
# encoding: utf-8

# esmre.py - clue-indexed regular expressions module
# Copyright (C) 2007-2008 Tideway Systems Limited.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA

import esm
import threading


class InBackslashState(object):
    def __init__(self, parent_state):
        self.parent_state = parent_state

    def process_byte(self, ch):
        return self.parent_state


class InClassState(object):
    def __init__(self, parent_state):
        self.parent_state = parent_state

    def process_byte(self, ch):
        if ch == "]":
            return self.parent_state

        elif ch == "\\":
            return InBackslashState(self)

        else:
            return self


class InBracesState(object):
    def __init__(self, parent_state):
        self.parent_state = parent_state

    def process_byte(self, ch):
        if ch == "}":
            return self.parent_state

        else:
            return self


class CollectingState(object):
    def __init__(self):
        self.hints = [""]

    def process_byte(self, ch):
        self.update_hints(ch)
        return self.next_state(ch)

    def bank_current_hint_with_last_byte(self):
        self.hints.append("")

    def bank_current_hint_and_forget_last_byte(self):
        if isinstance(self.hints[-1], list):
            del self.hints[-1]
        else:
            self.hints[-1] = self.hints[-1][:-1]

        self.hints.append("")

    def forget_all_hints(self):
        self.hints = [""]

    def append_to_current_hint(self, ch):
        self.hints[-1] += ch

    def update_hints(self, ch):
        if ch in "?*{":
            self.bank_current_hint_and_forget_last_byte()

        elif ch in "+.^$([\\":
            self.bank_current_hint_with_last_byte()

        elif ch == "|":
            self.forget_all_hints()

        else:
            self.append_to_current_hint(ch)

    def next_state(self, ch):
        if ch == "(":
            return StartOfGroupState(self)

        elif ch == "[":
            return InClassState(self)

        elif ch == "{":
            return InBracesState(self)

        elif ch == "\\":
            return InBackslashState(self)

        elif ch == "|":
            return self.alternation_state()

        else:
            return self

    def alternation_state(self):
        raise NotImplementedError


class RootState(CollectingState):
    def alternation_state(self):
        raise StopIteration


class StartOfGroupState(object):
    def __init__(self, parent_state):
        self.parent_state = parent_state

    def process_byte(self, ch):
        if ch == "?":
            return StartOfExtensionGroupState(self.parent_state)
        else:
            return InGroupState(self.parent_state).process_byte(ch)


class InGroupState(CollectingState):
    def __init__(self, parent_state):
        CollectingState.__init__(self)
        self.parent_state = parent_state
        self.had_alternation = False

    def update_hints(self, ch):
        if ch == ")":
            if not self.had_alternation:
                self.parent_state.hints.append(self.hints)
        else:
            CollectingState.update_hints(self, ch)

    def next_state(self, ch):
        if ch == ")":
            return self.close_group_state()
        else:
            return CollectingState.next_state(self, ch)

    def close_group_state(self):
        return self.parent_state

    def alternation_state(self):
        self.had_alternation = True
        return self


class StartOfExtensionGroupState(object):
    def __init__(self, parent_state):
        self.parent_state = parent_state

    def process_byte(self, ch):
        if ch == "P":
            return MaybeStartOfNamedGroupState(self.parent_state)
        else:
            return IgnoredGroupState(self.parent_state).process_byte(ch)


class MaybeStartOfNamedGroupState(object):
    def __init__(self, parent_state):
        self.parent_state = parent_state

    def process_byte(self, ch):
        if ch == "<":
            return InNamedGroupNameState(self.parent_state)
        else:
            return IgnoredGroupState(self.parent_state)


class InNamedGroupNameState(object):
    def __init__(self, parent_state):
        self.parent_state = parent_state

    def process_byte(self, ch):
        if ch == ">":
            return InGroupState(self.parent_state)
        else:
            return self


class IgnoredGroupState(InGroupState):
    def update_hints(self, ch):
        pass


def hints(regex):
    state = RootState()

    try:
        for ch in regex:
            state = state.process_byte(ch)

    except StopIteration:
        pass

    def flattened(hints):
        for item in hints:
            if isinstance(item, list):
                for i in flattened(item):
                    yield i
            else:
                yield item

    return [hint for hint in flattened(state.hints) if hint]


def shortlist(hints):
    if not hints:
        return []

    best = ""

    for hint in hints:
        if len(hint) > len(best):
            best = hint

    return [best]


class Index(object):
    def __init__(self):
        self.esm = esm.Index()
        self.hintless_objects = list()
        self.fixed = False
        self.lock = threading.Lock()

    def enter(self, regex, obj):
        self.lock.acquire()
        try:

            if self.fixed:
                raise TypeError("enter() cannot be called after query()")

            keywords = shortlist(hints(regex))

            if not keywords:
                self.hintless_objects.append(obj)

            for hint in shortlist(hints(regex)):
                self.esm.enter(hint.lower(), obj)

        finally:
            self.lock.release()

    def query(self, string):
        self.lock.acquire()
        try:

            if not self.fixed:
                self.esm.fix()
                self.fixed = True

        finally:
            self.lock.release()

        return self.hintless_objects + [
            obj for (_, obj) in self.esm.query(string.lower())
        ]



cdef extern from "esm/aho_corasick.h":
    ctypedef enum ac_error_code:
        AC_SUCCESS = 0
        AC_FAILURE = 1
    ctypedef ac_error_code (*ac_free_function)(void *item, void *data)
    ctypedef int ac_offset
    ctypedef struct ac_result:
        ac_offset start
        ac_offset end
        void *object
    ctypedef ac_error_code (*ac_result_callback)(void* result_cb_data,
                                                 ac_result* result)
    ctypedef struct ac_index_s:
        pass
    ctypedef ac_index_s* ac_index

    ac_index ac_index_new()
    ac_error_code ac_index_free(ac_index self,
                                ac_free_function object_free)

    ac_error_code ac_index_enter(ac_index self,
                                 char* keyword,
                                 ac_offset size,
                                 void* object)

    ac_error_code ac_index_fix(ac_index self)
    bint ac_index_fixed(ac_index self)

    ac_error_code ac_index_query_cb(ac_index self,
                                    char* phrase,
                                    ac_offset size,
                                    ac_result_callback result_cb,
                                    void* result_cb_data)

cdef extern from "Python.h":
    void Py_IncRef(obj)
    void Py_DecRef(obj)



#!/usr/bin/env python
# encoding: utf-8

# esm_tests.py - tests for esm extension module
# Copyright (C) 2007 Tideway Systems Limited.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA

import unittest
import esm


class IndexTests(unittest.TestCase):
    def testHasConstructor(self):
        self.assertTrue(esm.Index())

    def testIndexHasEnterMethod(self):
        esm.Index().enter("keyword")

    def testIndexMustHaveStringForFirstArgument(self):
        self.assertRaises(TypeError, esm.Index().enter, 0)
        self.assertRaises(TypeError, esm.Index().enter, [])
        self.assertRaises(TypeError, esm.Index().enter)

    def testIndexFix(self):
        index = esm.Index()
        index.fix()

    def testQuery(self):
        index = esm.Index()
        index.enter("he")
        index.enter("she")
        index.enter("his")
        index.enter("hers")
        index.fix()

        self.assertEqual(
            [((1, 4), "his"), ((5, 7), "he"), ((13, 16), "his")],
            index.query("this here is history"),
        )

    #                                     0123456789.123456789

    def testEnterObject(self):
        index = esm.Index()

        mint_object = dict()
        index.enter("mint", mint_object)

        pepper_object = dict()
        index.enter("pepper", pepper_object)

        index.fix()
        results = index.query("mint sauce")

        self.assertEqual(1, len(results))
        self.assertTrue(isinstance(results[0], tuple))

        slice_indices, associated_object = results[0]

        self.assertEqual((0, 4), slice_indices)
        self.assertTrue(associated_object is mint_object)

    def testCantFixIndexWhenAlreadyFixed(self):
        index = esm.Index()
        index.fix()

        self.assertRaises(TypeError, index.fix)

    def testCantEnterWhenAlreadyFixed(self):
        index = esm.Index()
        index.fix()

        self.assertRaises(TypeError, index.enter, "foo")

    def testQueryUntilFixed(self):
        index = esm.Index()
        self.assertRaises(TypeError, index.query, "foo")

    def testObjectsForCommonEndingsAreDecrefedCorrectly(self):
        o = "Owt"
        import sys

        initial_ref_count = sys.getrefcount(o)

        index = esm.Index()
        index.enter("food", o)
        index.enter("ood", o)
        index.fix()
        index.query("blah")

        index = None
        del index

        self.assertEqual(initial_ref_count, sys.getrefcount(o))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8

# esmre_tests.py - tests for esmre module
# Copyright (C) 2007-2008 Tideway Systems Limited.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA

import unittest
import esmre


class HintExtractionTests(unittest.TestCase):
    def checkHints(self, expected_hints, regex):
        self.assertEqual(set(expected_hints), set(esmre.hints(regex)))

    def testSimpleString(self):
        self.checkHints(["yarr"], r"yarr")

    def testSkipsOptionalCharacter(self):
        self.checkHints(["dubloon"], r"dubloons?")

    def testStartsNewStringAfterOptionalCharacter(self):
        self.checkHints(["ship", "shape"], r"ship ?shape")

    def testSkipsOptionalRepeatedCharacter(self):
        self.checkHints(["bristol", "fasion"], r"bristol *fasion")

    def testIncludesRepeatedCharacterButStartsNewHint(self):
        self.checkHints(["ava", "st me harties"], r"ava+st me harties")

    def testSkipsGroupsWithAlternation(self):
        self.checkHints(
            ["Hoist the ", ", ye ", "!"],
            r"Hoist the (mizzen mast|main brace), "
            r"ye (landlubbers|scurvy dogs)!",
        )

    def testSkipsAny(self):
        self.checkHints(
            ["After 10 paces, ", " marks the spot"],
            r"After 10 paces, . marks the spot",
        )

    def testSkipsOneOrMoreAny(self):
        self.checkHints(["Hard to ", "!"], r"Hard to .+!")

    def testSkipsNestedGroups(self):
        self.checkHints(
            ["Squark!", " Pieces of ", "!"], r"Squark!( Pieces of (.+)!)"
        )

    def testSkipsCharacterClass(self):
        self.checkHints(["r"], r"[ya]a*r+")

    def testRightBracketDoesNotCloseGroupIfInClass(self):
        self.checkHints([":=", "X"], r":=([)D])X")

    def testSkipsBackslashMetacharacters(self):
        self.checkHints(["Cap'n", " ", " Beard"], r"Cap'n\b ([\S] Beard)")

    def testBackslashBracketDoesNotCloseGroup(self):
        self.checkHints([":=", "X"], r":=(\)|D)X")

    def testBackslashSquareBracketDoesNotCloseClass(self):
        self.checkHints([":=", "X"], r":=[)D\]]X")

    def testSkipsMetacharactersAfterGroups(self):
        self.checkHints(
            ["Yo ", "ho ", " and a bottle of rum"],
            r"Yo (ho )+ and a bottle of rum",
        )

    def testSkipsRepetionBraces(self):
        self.checkHints(["A", ", me harties"], r"Ar{2-10}, me harties")

    def testAlternationCausesEmptyResult(self):
        self.checkHints([], r"rum|grog")

    def testSkipMatchBeginning(self):
        self.checkHints(["The black perl"], r"^The black perl")

    def testSkipMatchEnd(self):
        self.checkHints(["Davey Jones' Locker"], r"Davey Jones' Locker$")

    def testOnlyGroupGivesEmptyResult(self):
        self.checkHints([], r"(rum|grog)")

    def testGetsHintsFromGroups(self):
        self.checkHints(["/"], r"([0-3][0-9]/[0-1][0-9]/[1-2][0-9]{3})")

    def testSkipsOptionalGroups(self):
        self.checkHints(["Shiver me timbers!"], r"Shiver me timbers!( Arrr!)?")

    def testSkipsMostExtensionGroups(self):
        for regex in [
            # set flag
            r"(?i)(?L)(?m)(?s)(?u)(?x)",
            # non-grouping paren
            r"(?:foo)",
            # previous named group
            r"(?P=foo)",
            # comment
            r"(?#foo)",
            # lookahead
            r"(?=foo)",
            # negative lookahead
            r"(?!foo)",
            # lookbehind
            r"(?<=foo)",
            # negative lookbehind
            r"(?<!foo)",
            # conditional match
            r"(?(1)foo|bar)",
        ]:

            self.checkHints([], regex)

    def testGetsHintsFromNamedGroup(self):
        self.checkHints(
            ["/"], r"(?P<date>[0-3][0-9]/[0-1][0-9]/[1-2][0-9]{3})"
        )


class ShortlistTests(unittest.TestCase):
    def checkShortlist(self, expected_shortlist, hints):
        self.assertEqual(expected_shortlist, esmre.shortlist(hints))

    def testShortlistIsEmptyForEmptyCandidates(self):
        self.checkShortlist([], [])

    def testShortlistIsOnlyCandidate(self):
        self.checkShortlist(["Blue Beard"], ["Blue Beard"])

    def testShorlistSelectsLongestCandidate(self):
        self.checkShortlist(["Black Beard"], ["Black Beard", "Blue Beard"])

    def testShorlistSelectsLongestCandidateAtEnd(self):
        self.checkShortlist(
            ["Yellow Beard"], ["Black Beard", "Blue Beard", "Yellow Beard"]
        )


class IndexTests(unittest.TestCase):
    def setUp(self):
        self.index = esmre.Index()
        self.index.enter(r"Major-General\W*$", "savoy opera")
        self.index.enter(r"\bway\W+haye?\b", "sea shanty")

    def testSingleQuery(self):
        self.assertEqual(
            ["savoy opera"],
            self.index.query("I am the very model of a modern Major-General."),
        )

    def testCannotEnterAfterQuery(self):
        self.index.query("blah")
        self.assertRaises(TypeError, self.index.enter, "foo", "bar")

    def testCaseInsensitive(self):
        self.assertEqual(
            ["sea shanty"], self.index.query("Way, hay up she rises,")
        )
        self.assertEqual(
            ["sea shanty"],
            self.index.query("To my way haye, blow the man down,"),
        )

    def testAlwaysReportsOpjectForHintlessExpressions(self):
        self.index.enter(r"(\d+\s)*(paces|yards)", "distance")
        self.assertTrue("distance" in self.index.query("'til morning"))


if __name__ == "__main__":
    unittest.main()



# setup.py - distutils configuration for esm and esmre modules
# Copyright (C) 2007 Tideway Systems Limited.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA

from setuptools import setup, Extension
from Cython.Build import cythonize

module1 = Extension(
    "esm", ["src/esm.pyx", "src/esm/aho_corasick.c", "src/esm/ac_list.c"]
)

with open("README.md", "r") as readme:
    long_description = readme.read()

setup(
    name="esmre",
    version="1.0.1",
    description="Regular expression accelerator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: "
        "GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: POSIX",
        "Programming Language :: C",
        "Programming Language :: Cython",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Indexing",
    ],
    install_requires=["setuptools"],
    author="Will Harris",
    author_email="esmre@greatlibrary.net",
    url="https://github.com/wharris/esmre",
    license="GNU LGPL",
    platforms=["POSIX"],
    ext_modules=cythonize([module1]),
    package_dir={"": "src"},
    py_modules=["esmre"],
)
